class FacetsFilter {

    constructor(name = 'start') {
        this._resetFilterState(name);
        this._preFilter = null;
    }

    /* */

    checkOption(property, values) {
        if (!(values instanceof Array)) {
            values = [values];
        }
        if (!this._optionFilters[property]) {
            this._optionFilters[property] = [];
        }
        try {
            values.forEach((value) => {
                if (this._optionFilters[property].indexOf(value) === -1) {
                    this._optionFilters[property].push(value);
                }
            });
        }
        catch (e) {
            console.log('error occured: ', e);
        }
        return this;
    }

    clearOptions(property) {
        delete this._optionFilters[property];
        return this;
    }

    unCheckOption(property, value) {
        let index = this._optionFilters[property].indexOf(value);
        this._optionFilters[property].splice(index, 1);
        if (this._optionFilters[property].length === 0) {
            delete this._optionFilters[property];
        }
        return this;
    }

    clearAllOptions() {
        this._optionFilters = {};
        return this;
    }

    /* */

    setRange(property, min, max) {
        this._rangeFilters[property] = [min, max];
        return this;
    }

    clearRange(property) {
        delete this._rangeFilters[property];
        return this;
    }

    /* */

    setCustomFilter(property, filter, answer = null) {
        this._customFilters[property] = filter;
        this.setCustomFilterAnswer(property, answer);
        return this;
    }

    setCustomFilterAnswer(property, answer) {
        if (typeof this._customFilters[property] === 'undefined') {
            throw new Error('Cannot set answer for unknown custom filter: ' + property);
        }
        this._customFilterAnswers[property] = answer;
    }

    clearCustomFilter(property) {
        delete this._customFilters[property];
        delete this._customFilterAnswers[property];
        return this;
    }


    /* */

    getOptions(data, key, addKey = true) {
        return FacetsFilter.getOptionsFromData(this.execute(data), key, addKey);
    }

    static getOptionsFromData(data, key, addKey = true) {
        //if multiple keys are passed, merge results of each key.
        if (key instanceof Array) {
            return FacetsFilter.unique(key.reduce((previous, current) => {
                return previous.concat(FacetsFilter.getOptionsFromData(data, current, addKey));
            }, []));
        }


        let options = [];
        let values = [];
        let option;

        for (let i = 0; i < data.length; i++) {
            option = FacetsFilter.getValue(data[i], key);
            if (values.indexOf(option) === -1) {
                if (!FacetsFilter.isNull(option)) {
                    options.push({name: option, parent: key});
                    values.push(option);
                }
            }
        }
        return addKey ? options : values;
    }


    getActiveOptions(key, includePreFilter = true) {
        if (this._preFilter && includePreFilter) {
            return FacetsFilter.unique(this._preFilter.getActiveOptions(key, includePreFilter).concat(this._optionFilters[key] || []));
        }

        return this._optionFilters[key] || [];
    }

    getCustomFilterAnswer(key, includePreFilter = true) {
        if (this._preFilter && includePreFilter) {
            return this._customFilterAnswers[key] || this._preFilter.getCustomFilterAnswer(key, includePreFilter);
        }

        return this._customFilterAnswers[key] || null;
    }

    assertStage(name, restoreIfExists = true) {
        console.log('AssertStage ', name, restoreIfExists);
        if (restoreIfExists && this.hasStage(name)) {
            console.log('Restoring stage: ', name);
            this.restoreStage(name);
        }

        if (!this.hasStage(name)) {
            console.log('Creating new stage: ', name);
            const preFilter = new FacetsFilter();
            preFilter._restoreFilterState(this._extractFilterState());
            preFilter._preFilter = this._preFilter;

            this._name = name;
            this._preFilter = preFilter;
        }
    }

    stage(name, throwErrorOnMissing = true) {

        const result = this._getPreFilterStage(name);

        if (result === null && throwErrorOnMissing) {
            throw new Error('Stage ' + name + ' does not exist');
        }

        return result || this;

    }

    _getPreFilterStage(name) {
        if (this._name === name) {
            return this;
        }
        if (this._preFilter) {
            return this._preFilter._getPreFilterStage(name);
        }
        return null;
    }

    hasStage(name) {
        if (this._name === name) {
            return true;
        }
        if (this._preFilter) {
            return this._preFilter.hasStage(name);
        }
        return false;
    }

    restoreStage(name) {
        if (this._name === name) {
            console.log('Active stage already is: ', name, '. Skipping restore');
        }
        while (this.hasStage(name) && this._name !== name) {
            if (this._preFilter) {
                this._restoreFilterState(this._preFilter._extractFilterState());
                this._preFilter = this._preFilter._preFilter;
            }
        }
    }

    /* */

    printDebugOutput(data, showFullData = false, showOriginalData = true, successorFields = []) {
        const uniqueFields = FacetsFilter.unique(this._keys(this._optionFilters)
            .concat(this._keys(this._rangeFilters))
            .concat(this._keys(this._customFilters))
            .concat(successorFields)
        );
        if (showOriginalData) {
            console.log('Original data: ', showFullData ? data : data.length + ' item(s)');
        }
        const preFilterFields = this._preFilter ? this._preFilter.printDebugOutput(data, showFullData, false, uniqueFields) : uniqueFields;


        console.log('=== Stage: ', this._name, " ===");

        this._keys(this._optionFilters).forEach(k => {
            console.log('  Option filter: ', k, this._optionFilters[k]);
        });

        this._keys(this._rangeFilters).forEach(k => {
            console.log('  Range filter: ', k, this._rangeFilters[k]);
        });

        if (this._keys(this._customFilters).length > 0) {
            console.log('  Custom filters: ', this._keys(this._customFilters).join("; "));
        }

        const filteredData = this.execute(data),
            remainingOptions = {},
            customFilterAnswers = {};

        console.log('  > Filtered data: ', showFullData ? filteredData : filteredData.length + ' item(s)');

        preFilterFields.forEach(k => {

            remainingOptions[k] = FacetsFilter.unique(this.getOptions(filteredData, k, false));
            if (remainingOptions[k].length < 2) {
                remainingOptions[k] = remainingOptions[k][0];
            }

            const customFilterAnswer = this.getCustomFilterAnswer(k);
            if (!FacetsFilter.isNull(customFilterAnswer)) {
                customFilterAnswers[k] = customFilterAnswer;
            }

        });

        console.log('  > Custom filter answers: ', customFilterAnswers);
        console.log('  > Remaining options: ', remainingOptions);

        return preFilterFields;
    }

    _keys(obj) {
        return Object.keys(obj).filter(k => obj.hasOwnProperty(k));
    }

    /* */
    /* eslint-disable */
    _extractFilterState() {
        const result = {
            name: this._name,
            optionFilters: JSON.parse(JSON.stringify(this._optionFilters)),
            rangeFilters: JSON.parse(JSON.stringify(this._rangeFilters)),
            customFilters: object.assign({}, this._customFilters),
            customFilterAnswers: JSON.parse(JSON.stringify(this._customFilterAnswers))
        };

        this._resetFilterState();

        return result;
    }
    /* eslint-enable */

    _restoreFilterState(state) {
        this._name = state.name;
        this._optionFilters = state.optionFilters;
        this._rangeFilters = state.rangeFilters;
        this._customFilters = state.customFilters;
        this._customFilterAnswers = state.customFilterAnswers;
    }

    _resetFilterState(name) {
        this._name = name;
        this._optionFilters = {};
        this._rangeFilters = {};
        this._customFilters = {};
        this._customFilterAnswers = {};
    }

    _preFilterData(data, includeNullValues = false) {
        return this._preFilter ? this._preFilter.execute(data, includeNullValues) : data;
    }

    /* */

    execute(data, includeNullValues = false) {
        //console.warn('EXPENSIVE FILTER METHOD CALLED');

        let result = this._preFilterData(data, includeNullValues);

        Object.keys(this._optionFilters).forEach(key => {
            result = result.filter((item) => {
                let value = FacetsFilter.getValue(item, key);
                try {
                    return (
                        (includeNullValues && FacetsFilter.isNull(value)) ||
                        this._optionFilters[key].indexOf(value) > -1
                    );
                }
                catch (e) {
                    console.log('execute error occured: ', e);
                }
            });
        });

        Object.keys(this._customFilters).forEach(key => {
            result = result.filter((item) => {
                let value = FacetsFilter.getValue(item, key);
                return this._customFilters[key](value, this._customFilterAnswers[key]);
            });
        });

        Object.keys(this._rangeFilters).forEach(key => {
            result = result.filter((item) => {
                let value = FacetsFilter.getValue(item, key);
                return (
                    (includeNullValues && FacetsFilter.isNull(value)) ||
                    (
                        value >= this._rangeFilters[key][0] &&
                        value <= this._rangeFilters[key][1]
                    )
                );
            });
        });

        return result;
    }

    static getValue(object, key) {
        let path = key.split('.');
        let result = object;

        path.forEach(property => {
            if (typeof result === 'object') {
                result = result[property];
            }
        });

        return result;
    }

    static isNull(value) {
        return (typeof value === 'undefined' || value === null || value === '');
    }

    static unique(array) {
        const values = {};
        return array.filter((item) => {
            if (typeof item === 'object') {
                item = JSON.stringify(item);
            }

            const result = (typeof values[item]) === 'undefined';
            values[item] = true;

            return result;
        });
    }
}

class CachedFacetsFilter {

    constructor(facetsFilter, data, includeNullValues = false) {
        this._originalData = data;
        this._filteredData = null;
        this._includeNullValues = includeNullValues;
        this._dirty = true;
        this._filter = facetsFilter;
    }

    checkOption(property, values) {
        this._filter.checkOption(property, values);

        this._flagDirty();
        return this;
    }

    setCustomFilter(property) {
        this._filter.setCustomFilter(property);
        return this;
    }

    clearOptions(property) {
        this._filter.clearOptions(property);

        this._flagDirty();
        return this;
    }

    setRange(property, min, max) {
        this._filter.setRange[property] = [min, max];
        return this;
    }

    clearRange(property) {
        this._filter.clearRange(property);
        return this;
    }

    execute() {
        if (this._dirty) {
            this._updateResult();
        }

        return this._filteredData;
    }

    _flagDirty() {
        this._dirty = true;
    }

    _updateResult() {
        this._filteredData = this._filter.execute(this._originalData, this._includeNullValues);
        this._dirty = false;
    }
}

export {
    FacetsFilter,
    CachedFacetsFilter
};
